require 'runit/version'
require 'runit/error'

module RUNIT
  module Assert
    FAILURE_SKIP = [false]

    def Assert.skip_failure=(arg)
      FAILURE_SKIP[0] = arg
    end

    def Assert.skip_failure?
      FAILURE_SKIP[0]
    end

    def edit_message(msg)
      if msg != ""
        msg.concat " "
      end
      msg
    end
    private :edit_message

    def to_str(obj)
      if obj.kind_of?(String)
        obj
      else
        obj.inspect
      end
    end
    private :to_str

    def to_str_with_type(obj)
      "<" + to_str(obj) + ":" + to_str(obj.type) + ">"
    end
    private :to_str_with_type

    def setup_assert
    end
    private :setup_assert

    def assert(condition, message="")
      setup_assert
      unless condition
	msg = edit_message(message)
	msg.concat 'The condition is '
	msg.concat to_str_with_type(condition)
        raise_assertion_error(msg, 2)
      end
    end

    def assert_equal(expected, actual, message="")
      setup_assert
      if expected != actual
        msg = edit_message(message)
        msg.concat "expected:<"
	msg.concat to_str(expected)
        msg.concat "> but was:<"
        msg.concat to_str(actual)
        msg.concat ">"
        raise_assertion_error(msg, 2) 
      end
    end
    alias assert_equals assert_equal

    def assert_operator(obj1, op, obj2, message="")
      setup_assert
      if !obj1.send(op, obj2)
        msg = edit_message(message)
        msg.concat "The condition is not satisfied: "
	msg.concat to_str(obj1)
        msg.concat " "
	msg.concat op.to_s
        msg.concat " "
	msg.concat to_str(obj2)
        raise_assertion_error(msg, 2) 
      end
    end

    def assert_equal_float(expected, actual, e, message="")
      if e < 0.0
        raise ArgumentError, "#{e}: 3rd argument should be 0 or greater than 0."
      end
      if e == 0.0
	return assert_equal(expected, actual, message)
      end
      setup_assert
      if (expected - actual).abs > e
	i = 0
	while 1 > e
	  i += 1
	  e *= 10
	end
	i += 1

	msg = edit_message(message)
	msg.concat "expected:<"
	msg.concat sprintf("%.#{i}f", expected)
        msg.concat "> but was:<"
	msg.concat sprintf("%.#{i}f", actual)
        msg.concat ">"
        raise_assertion_error(msg, 2) 
      end
    end

    def assert_same(expected, actual, message="")
      setup_assert
      if !actual.equal?(expected)
	msg = edit_message(message)
        msg.concat to_str_with_type(actual) 
	msg.concat " is not same object: "
	msg.concat to_str_with_type(expected) 
	raise_assertion_error(msg, 2)
      end
    end

    def assert_send(obj, method, *args)
      setup_assert
      if !obj.send(method, *args)
        msg = "Assertion failed: "
        msg.concat obj.type.to_s
        msg.concat "("
	msg.concat obj.inspect
        msg.concat ")"
        msg.concat "#"
	msg.concat method.to_s
        if args.size > 0 
		  msg.concat "("
          msg.concat args.inspect
		  msg.concat ")"
        end
        raise_assertion_error(msg, 2) 
      end
    end

    def assert_nil(obj, message="")
      setup_assert
      if !obj.nil?
	msg = edit_message(message)
        msg += to_str_with_type(obj) + " is not nil"
	raise_assertion_error(msg, 2)
      end
    end

    def assert_not_nil(obj, message="")
      setup_assert
      if obj.nil?
	msg = edit_message(message)
	msg.concat "object is nil"
	raise_assertion_error(msg, 2)
      end
    end

    def assert_respond_to(method, obj, message="")
      setup_assert
      if !obj.respond_to?(method)
	msg = edit_message(message)
        method_str = method
        if !method.instance_of?(String)
          method_str = method.id2name
        end
	msg += to_str_with_type(obj)
	msg += " does not respond to <" + method_str + ">"
	raise_assertion_error(msg, 2)
      end
    end

    def assert_kind_of(c, obj, message="")
      setup_assert
      if !obj.kind_of?(c)
	msg = edit_message(message)
	msg.concat to_str_with_type(obj)
	msg.concat " is not kind of <" + c.to_s + ">"
	raise_assertion_error(msg, 2)
      end
    end

    def assert_instance_of(c, obj, message="")
      setup_assert
      if !obj.instance_of?(c)
	msg = edit_message(message)
	msg.concat to_str_with_type(obj)
	msg.concat " is not instance of <" + c.to_s + ">"
	raise_assertion_error(msg, 2)
      end
    end

    def assert_match(str, re, message="")
      setup_assert
      if re !~ str
	msg = edit_message(message)
        r = to_str(re)
	s = to_str(str)
	msg.concat ("<" + r + "> not match <" + s + ">")
	raise_assertion_error(msg, 2)
      end
    end
    alias assert_matches assert_match

    def assert_not_match(str, re, message="")
      setup_assert
      if re =~ str
	msg = edit_message(message)
	r = to_str(re)
	s = to_str(str)
	msg.concat ("<" + r + "> matches <" + s + ">")
	raise_assertion_error(msg, 2)
      end
    end

    def assert_exception(exception, message="")
      setup_assert
      block = proc
      exception_raised = true
      err = ""
      ret = nil
      begin
	block.call
	exception_raised = false
	err = 'NO EXCEPTION RAISED'
      rescue Exception
	if $!.instance_of?(exception)
	  exception_raised = true
	  ret = $!
	else
	  raise $!.type, $!.message, $!.backtrace
	end
      end
      if !exception_raised
      	msg = edit_message(message)
        msg.concat "expected:<"
	msg.concat to_str(exception)
	msg.concat "> but was:<"
	msg.concat to_str(err)
	msg.concat ">"
	raise_assertion_error(msg, 2)
      end
      ret
    end

    def assert_no_exception(*arg)
      setup_assert
      block = proc
      message = ""
      if arg[arg.size-1].instance_of?(String)
	message = arg.pop
      end
      err = nil
      exception_raised = false
      begin
	block.call
      rescue Exception
	if arg.include?($!.type) || arg.size == 0
	  exception_raised = true
	  err = $!
        else
	  raise $!.type, $!.message, $!.backtrace
	end
      end
      if exception_raised
      	msg = edit_message(message)
	msg.concat "Exception raised:"
	msg.concat to_str(err)
	raise_assertion_error(msg, 2)
      end
    end

    def assert_fail(message, level=2)
      setup_assert
      raise_assertion_error message, level
    end

    def raise_assertion_error(message, level=2)
      if Assert.skip_failure?
	changed
	notify_observers(RObserver::ADD_FAILURE, caller(level), message, type) 
      else
        raise AssertionFailedError, message, caller(level)
      end
    end
    private :raise_assertion_error

  end
end
